# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import importlib
import warnings
from autologging import logged

from kqcircuits.elements.element import insert_cell_into
from kqcircuits.pya_resolver import pya, lay, is_standalone_session

from kqcircuits.defaults import default_layers, default_png_dimensions, mask_bitmap_export_layers, \
    all_layers_bitmap_hide_layers, default_faces, default_layer_props


@logged
class KLayoutView:
    """Helper object to represent the KLayout rendering environment.

    ``KLayoutView`` is a wrapper around the KLayout ``LayoutView`` and ``CellView`` objects, that represent containers
    for viewing a layout in the GUI. It provides methods to initialize the views and layout for KQCircuits, for placing
    KQCircuits ``Elements``, and for exporting images.

    Create a new view as follows::

       view = KLayoutView()

    This creates a new set of ``LayoutView``, ``CellView``, ``Layout`` and top ``Cell`` objects with layers initialized
    to the KQCircuits layer configuration. In the KLayout GUI, the new view will show as a new tab.

    Note: In standalone python mode, the user must keep a reference to the ``KLayoutView`` object in scope, otherwise
    the associated layout and cells may also go out of scope.

    When running scripts or macros in the KLayout application, the following command creates a wrapper around the
    currently active view::

        view = KLayoutView(current=True)

    This can be used in macros to act on the existing layout. The argument ``current=True`` is not available in
    standalone python mode.

    Once a view is created, new ``Elements`` can be placed with the ``insert_cell`` method.

    Several methods are available to export PNG files of the current view or specific cells and layers.
    In Jupyter notebooks, the ``show`` method displays the current view inline in the notebook.
    """
    if hasattr(lay, "LayoutView"):
        layout_view: lay.LayoutView

    def __init__(self, current=False, initialize=None, background_color="#ffffff"):
        """Initialize a ``KLayoutView`` instance.

        Args:
            current: Boolean. If True, wrap the currently active ``LayoutView`` (not available in standalone python)
            initialize: Boolean, specify whether to initialize the layout with the default layer configuration and a
                top cell. Defaults to True if ``current==False``, and to False if ``current==True``.
            background_color: Background color as HTML color code. Defaults to `"#ffffff"` (white).
            """
        if not hasattr(lay, "LayoutView"):
            # Standalone session before KLayout 0.28
            raise MissingUILibraryException("KLayoutView is not supported in standalone mode for this klayout version. "
                                            "Consider upgrading your klayout package to version 0.28 or above.")

        if initialize is None:
            initialize = not current

        if is_standalone_session():
            # Standalone session since KLayout 0.28
            if current:
                raise MissingUILibraryException("In standalone python mode only KLayoutView(current=False) " +
                                                "is supported.")
            self.layout_view = lay.LayoutView(True)  # Creates a new LayoutView in editable mode
            self.layout_view.show_layout(pya.Layout(), True)  # Adds a CellView and Layout
            self.layout_view.set_config("background-color", background_color)
        else:
            # Regular session in the KLayout GUI
            if not current or (lay.LayoutView.current() is None):
                pya.MainWindow.instance().create_layout(1)
            self.layout_view = lay.LayoutView.current()

        if initialize:
            self.add_default_layers()
            self.create_top_cell()

    def insert_cell(self, cell, trans=None, inst_name=None, label_trans=None, align_to=None, align=None,
                    rec_levels=0, **parameters):
        """Inserts a subcell into the first top cell (the very first cell in the cell window)

        It will use the given ``cell`` object or if ``cell`` is an Element class' name then directly
        take the provided keyword arguments to first create the cell object.

        If `inst_name` given, a label ``inst_name`` is added to labels layer at the ``base`` refpoint and `label_trans`
        transformation.

        Args:
            cell: cell object or Element class name
            trans: used transformation for placement. None by default, which places the subcell into the coordinate
                origin of the parent cell. If `align` and `align_to` arguments are used, `trans` is applied to the
                `cell` before alignment transform which allows for example rotation of the `cell` before placement.
            inst_name: possible instance name inserted into subcell properties under `id`. Default is None
            label_trans: relative transformation for the instance name label
            align_to: ``DPoint`` or ``DVector`` location in parent cell coordinates for alignment of cell.
                Default is None
            align: name of the ``cell`` refpoint aligned to argument ``align_to``. Default is None
            rec_levels: recursion level when looking for refpoints from subcells. Set to 0 to disable recursion.
            **parameters: PCell parameters for the element, as keyword argument

        Return:
            tuple of placed cell instance and reference points with the same transformation
        """

        return insert_cell_into(self.top_cell, cell, trans, inst_name, label_trans, align_to, align, rec_levels,
                                **parameters)

    def focus(self, cell=None):
        """Sets a given cell as the active cell, and fits the zoom level to fit the cell.

        Args:
            cell: cell to focus on, or None to focus on the currently active cell
        """
        if cell is not None and isinstance(cell, pya.Cell):
            self.layout_view.active_cellview().active().cell = cell
        self.layout_view.max_hier()
        self.layout_view.zoom_fit()

    def show(self, **kwargs):
        """In KLayout, show this LayoutView as the current in the main window.

        In standalone python, display an image of the view. Requires IPython / Jupyter. Keyword arguments are passed
        to ``KLayoutView.get_pixels``.
        """

        if is_standalone_session():
            display = importlib.import_module("IPython.display")
            display.display_png(self.get_pixels(**kwargs).to_png_data(), raw=True)
        else:
            main_window = lay.MainWindow.instance()
            for i in range(main_window.views()):
                if main_window.view(i) is self.layout_view:
                    main_window.current_view_index = i
                    break

    def close(self):
        """Closes the current LayoutView."""
        if is_standalone_session():
            self.layout_view.destroy()
        else:
            lay.MainWindow.instance().close_current_view()

    @property
    def cell_view(self) -> "lay.CellView":
        """The active ``CellView``"""
        return self.layout_view.active_cellview()

    @property
    def layout(self) -> pya.Layout:
        """The active ``Layout``"""
        return self.cell_view.layout()

    @property
    def active_cell(self) -> pya.Cell:
        """The active ``Cell``, which is shown as current top in the cellview and bold in the cell window.

        Can be set to any Cell in the layout."""
        return self.cell_view.cell

    @active_cell.setter
    def active_cell(self, cell: pya.Cell):
        self.cell_view.cell = cell

    @property
    def top_cell(self) -> pya.Cell:
        """The first top cell of the active layout."""
        cells = self.layout.top_cells()
        return cells[0] if len(cells) else None

    def clear_layers(self):
        """Clear the layer view."""
        self.layout_view.clear_layers()

    def add_default_layers(self):
        """Populate view with KQCircuits default layers. Adds the layers to the layout, and populates the layer view."""
        self.clear_layers()
        layout = self.layout
        for layer in default_layers.values():
            layout.layer(layer)
        self.layout_view.add_missing_layers()
        if default_layer_props:
            self.layout_view.load_layer_props(default_layer_props, True)

    def create_top_cell(self, top_cell_name="Top Cell"):
        """Creates a new static cell and set it as the top cell."""
        top_cell = self.layout.create_cell(top_cell_name)
        self.cell_view.cell_name = top_cell.name  # Shows the new cell
        return top_cell

    def export_layers_bitmaps(self, path, cell, filename=None, layers_set=None, face_id=None):
        """Exports each layer to a separate png image.

        Args:
            path: Directory to place the exported file in
            cell: Cell to export
            filename: Filename to export to, or None to use the cell's name.
            layers_set: A list of layer names to export, or None for default values specified in the layer configuration
            face_id: The face id for which to export the given layers, or None for general layers not associated to
                a face.
        """
        if layers_set is None:
            layers_set = mask_bitmap_export_layers
        for layer_name in layers_set:
            layer_info = resolve_default_layer_info(layer_name, face_id)
            self._export_bitmap(path, cell, filename=filename, layers_set=[layer_info])

    def export_all_layers_bitmap(self, path, cell, filename=None):
        """Exports a cell to a .png file with all layers visible

        Args:
            path: Directory to place the exported file in
            cell: Cell to export
            filename: Filename to export to, or None to use the cell's name.
        """
        self._export_bitmap(path, cell, filename=filename, layers_set='all')

    def export_pcell_png(self, path, cell, filename=None, max_size=default_png_dimensions[0]):
        """Exports a cell to a .png file no bigger than max_size at either dimension.

        Args:
            path: Directory to place the exported file in
            cell: Cell to export
            filename: Filename to export to, or None to use the cell's name.
            max_size: Maximum size of the image.
        """

        zoom = cell.dbbox()
        x, y = zoom.width(), zoom.height()
        if max_size * x / y < max_size - 200 :   # 200x100 is enough for the sizebar
            size = (max_size * x / y + 200, max_size)
        else:
            size = (max_size, max_size * y / x + 100)
        self._export_bitmap(path, cell, filename=filename, layers_set='all', z_box=zoom, pngsize=size)

    def get_pixels(self, cell=None, width=None, height=None, layers_set=None, box=None):
        """Returns a PixelBuffer render of the current view.

        This method first zooms to fit the whole layout and shows all hierarchy levels. If either ``width`` or
        ``height`` is specified, the other is chosen correspondingly to keep the same viewport aspect ratio. If neither
        is specified, the current viewport size is used.

        Args:
            cell: Cell to render, or None to render the currently active cell.
            width: image width in pixels, or None for automatic.
            height: image height in pixels, or None for automatic.
            layers_set: list of layer names to export, or None for the default set.
            box: DBox area to show, or None for the full layout.

        Returns: PixelBuffer
        """

        if width is None and height is None:
            width = self.layout_view.viewport_width()
            height = self.layout_view.viewport_height()
        elif width is None:
            width = height * self.layout_view.viewport_width() / self.layout_view.viewport_height()
        elif height is None:
            height = width * self.layout_view.viewport_height() / self.layout_view.viewport_width()
        if layers_set is not None:
            layers_set = [resolve_default_layer_info(layer_name) for layer_name in layers_set]

        def export_callback():
            pixel_buffer = self.layout_view.get_pixels(width, height)
            return pixel_buffer

        return self._export_bitmap_configure(export_callback, cell, layers_set, box)

    @staticmethod
    def get_active_cell_view():
        """Gets the currently active CellView. Not supported in standalone python mode.

        Deprecated, use ``KLayoutView(current=True).cell_view`` to get the same behavior."""
        warnings.warn('KLayoutView.get_active_cell_view will be deprecated. ' +
                      'Use instance property KLayoutView.cell_view instead.', DeprecationWarning)
        return lay.CellView.active()

    @staticmethod
    def get_active_layout():
        """Gets the layout of the currently active CellView. Not supported in standalone python mode.

        Deprecated, use ``KLayoutView(current=True).layout`` to get the same behavior. If you already have a
        KLayoutView instance, use the ``layout`` property of that instance instead."""
        warnings.warn('KLayoutView.get_active_layout will be deprecated. ' +
                      'Use instance property KLayoutView.layout instead.', DeprecationWarning)
        return lay.CellView.active().layout()

    @staticmethod
    def get_active_cell():
        """Gets the active cell of the currently active CellView. Not supported in standalone python mode.

        Deprecated, use ``KLayoutView(current=True).active_cell`` to get the same behavior. If you already have a
        KLayoutView instance, use the ``active_cell`` property of that instance instead."""
        warnings.warn('KLayoutView.get_active_cell will be deprecated. ' +
                      'Use instance property KLayoutView.active_cell instead.', DeprecationWarning)
        return lay.CellView.active().cell

    # ********************************************************************************
    # PRIVATE METHODS
    # ********************************************************************************

    def _export_bitmap(self, path, cell=None, filename=None, layers_set=None, z_box=None,
                       pngsize=default_png_dimensions):

        if filename is None:
            filename = cell.name
        if layers_set is None:
            layers_set = [resolve_default_layer_info(layer_name) for layer_name in mask_bitmap_export_layers]

        if len(layers_set) == 1:
            layer_str = "-" + layers_set[0].name
        else:
            layer_str = ""
        cell_png_name = path / "{}{}.png".format(filename, layer_str)

        def export_callback():
            self.layout_view.save_image(str(cell_png_name), pngsize[0], pngsize[1])

        self._export_bitmap_configure(export_callback, cell, layers_set, z_box)

    def _export_bitmap_configure(self, export_callback, cell, layers_set, z_box):
        """ Common configuration for export functions. """

        def get_visibility_state():
            """Get the current layer visibility and drawing focus state"""
            current_layer_visibility = [_layer.visible for _layer in self.layout_view.each_layer()]
            current_cell = self.layout_view.active_cellview().cell
            current_hier = (self.layout_view.min_hier_levels, self.layout_view.max_hier_levels)
            current_zoom = self.layout_view.box()
            return current_layer_visibility, current_cell, current_zoom, current_hier

        def restore_visibility_state(layer_visibility, cell, zoom, hier):
            """Restore the layer visibility and drawing focus state.
            Assumes order of layers has not changed since calling ``get_visibility_state`` """
            for _layer, _visible in zip(self.layout_view.each_layer(), layer_visibility):
                _layer.visible = _visible
            self.layout_view.active_cellview().cell = cell
            self.layout_view.min_hier_levels, self.layout_view.max_hier_levels = hier
            self.layout_view.zoom_box(zoom)

        visibility_state = get_visibility_state()

        if cell is not None:
            self.layout_view.active_cellview().cell = cell

        self.layout_view.max_hier()
        self.layout_view.zoom_fit()  # Has to be done also before zoom_box

        if z_box is not None:
            self.layout_view.zoom_box(z_box)

        if layers_set is None or layers_set == "all":
            for layer in self.layout_view.each_layer():
                layer.visible = True

                for layer_to_hide in all_layers_bitmap_hide_layers:
                    if layer.source_layer == layer_to_hide.layer and layer.source_datatype == layer_to_hide.datatype:
                        layer.visible = False
                        break
        else:
            layer_infos = self.layout.layer_infos()
            for layer in self.layout_view.each_layer():
                # need to avoid hiding layer groups, because that could hide also layers in layers_set
                is_layer_group = True
                for layer_info in layer_infos:
                    if pya.LayerInfo(layer.source_layer, layer.source_datatype).is_equivalent(layer_info):
                        is_layer_group = False
                        break
                if not is_layer_group:
                    layer.visible = False
                for layer_to_show in layers_set:
                    if layer.source_layer == layer_to_show.layer \
                            and layer.source_datatype == layer_to_show.datatype:
                        layer.visible = True
                        break

        if hasattr(lay, "Application"):
            # Make sure the layer property changes are reflected before saving the image
            lay.Application.instance().process_events()
        else:
            # Undocumented method timer does basically the same thing as process_events in GUI.
            self.layout_view.timer()

        export_return = export_callback()

        restore_visibility_state(*visibility_state)

        return export_return


class MissingUILibraryException(Exception):
    def __init__(self, message="Missing KLayout UI library."):
        Exception.__init__(self, message)


@logged
def resolve_default_layer_info(layer_name, face_id=None):
    """Returns LayerInfo based on default_layers.

    Assumes that layer_name is valid, and that face_id is valid or None.

    Args:
        layer_name: layer name (without face prefix for face-specific layers)
        face_id: id of the face from which this layer should be
    """
    if face_id:
        if layer_name in default_faces[face_id]:
            return default_faces[face_id][layer_name]
        else:
            return default_layers[layer_name]
    else:
        return default_layers[layer_name]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.defaults import default_marker_type


@logged
class Marker(Element):
    """Base Class for Markers."""

    default_type = default_marker_type

    diagonal_squares = Param(pdt.TypeInt, "Number of diagonal squares in the marker", 10)
    window = Param(pdt.TypeBoolean, "Window in airbridge flyover layer", False)

    @classmethod
    def create(cls, layout, library=None, marker_type=None, **parameters):
        """Create a Marker cell in layout."""
        return cls.create_subtype(layout, library, marker_type, **parameters)[0]

    def produce_geometry(self):
        """Produce common marker geometry."""

        layer_gap = self.get_layer("base_metal_gap_wo_grid")
        layer_pads = self.get_layer("airbridge_pads")
        layer_flyover = self.get_layer("airbridge_flyover")
        layer_gap_for_ebl = self.get_layer("base_metal_gap_for_EBL")
        layer_protection = self.get_layer("ground_grid_avoidance")

        def insert_to_main_layers(shape):
            self.cell.shapes(layer_gap).insert(shape)
            self.cell.shapes(layer_gap_for_ebl).insert(shape)
            if not self.window:
                self.cell.shapes(layer_flyover).insert(shape)

        # protection for the box
        protection_box = pya.DBox(
            pya.DPoint(220, 220),
            pya.DPoint(-220, -220)
        )
        self.cell.shapes(layer_protection).insert(protection_box)

        # make corners
        corner = pya.DPolygon([
            pya.DPoint(100, 100),
            pya.DPoint(10, 100),
            pya.DPoint(10, 80),
            pya.DPoint(80, 80),
            pya.DPoint(80, 10),
            pya.DPoint(100, 10),
        ])
        inner_corners = [pya.DTrans(a) * corner for a in [0, 1, 2, 3]]
        outer_corners = [pya.DCplxTrans(2, a * 90., False, pya.DVector()) * corner for a in [0, 1, 2, 3]]
        corners = pya.Region([s.to_itype(self.layout.dbu) for s in inner_corners + outer_corners])
        insert_to_main_layers(corners)

        # center box
        sqr_uni = pya.DBox(
            pya.DPoint(10, 10),
            pya.DPoint(-10, -10),
        )
        insert_to_main_layers(sqr_uni)

        self.inv_corners = pya.Region([protection_box.to_itype(self.layout.dbu)])
        self.inv_corners -= corners
        self.cell.shapes(layer_pads).insert(self.inv_corners - pya.Region(sqr_uni.to_itype(self.layout.dbu)))

        # window for airbridge flyover layer
        aflw = pya.DPolygon([
            pya.DPoint(800, 800),
            pya.DPoint(800, 10),
            pya.DPoint(80, 10),
            pya.DPoint(80, 2),
            pya.DPoint(2, 2),
            pya.DPoint(2, 80),
            pya.DPoint(10, 80),
            pya.DPoint(10, 800)
        ])
        if self.window:
            for alpha in [0, 1, 2, 3]:
                self.cell.shapes(layer_flyover).insert(pya.DTrans(alpha) * aflw)

        # marker diagonal
        sqr = pya.DBox(
            pya.DPoint(10, 10),
            pya.DPoint(2, 2),
        )
        self.diagonals = pya.Region()
        for i in range(5, 5 + self.diagonal_squares):
            ds = pya.DCplxTrans(3, 0, False, pya.DVector(50 * i - 3 * 6, 50 * i - 3 * 6)) * sqr
            insert_to_main_layers(ds)
            self.cell.shapes(layer_pads).insert(ds)
            self.diagonals += ds.to_itype(self.layout.dbu)
            self.cell.shapes(layer_protection).insert(
                pya.DCplxTrans(20, 0, False, pya.DVector(50 * i - 20 * 6, 50 * i - 20 * 6)) * sqr)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import numpy

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip


class JunctionTest(Chip):
    """The PCell declaration for a JunctionTest chip."""

    edge_len = Param(pdt.TypeInt, "Length of square's one edge", 80)
    inter_space = Param(pdt.TypeInt, "Space in between the Squares", 20)

    def build(self):

        # defining the parameters for local use

        left = self.box.left
        right = self.box.right
        top = self.box.top
        bottom = self.box.bottom

        width = right - left
        height = top - bottom
        dice_width = float(self.frames_dice_width[0])

        # create the polygon
        poly = pya.DPolygon([
            pya.DPoint(left + 2000, bottom + 100 + dice_width),
            pya.DPoint(left + 2000, bottom + 2000),
            pya.DPoint(left + 100 + dice_width, bottom + 2000),
            pya.DPoint(left + 100 + dice_width, bottom + 8000),
            pya.DPoint(left + 2000, bottom + 8000),
            pya.DPoint(left + 2000, bottom + height - 100 - dice_width),
            pya.DPoint(left + 8000, bottom + height - 100 - dice_width),
            pya.DPoint(left + 8000, bottom + 8000),
            pya.DPoint(left + width - 100 - dice_width, bottom + 8000),
            pya.DPoint(left + width - 100 - dice_width, bottom + 2000),
            pya.DPoint(left + 8000, bottom + 2000),
            pya.DPoint(left + 8000, bottom + 100 + dice_width)
        ])

        # create the box array
        b_array = []

        for y in numpy.arange(bottom + dice_width + self.inter_space, bottom + height - dice_width,
                              self.edge_len + self.inter_space):
            for x in numpy.arange(left + dice_width + self.inter_space, left + width - dice_width,
                                  self.edge_len + self.inter_space):
                b = pya.DPolygon(pya.DBox(x, y, x + self.edge_len, y + self.edge_len)).to_itype(self.layout.dbu)
                b_array.append(b)

        # substract the box array from the polygon
        reg1 = pya.Region(poly.to_itype(self.layout.dbu))
        reg2 = pya.Region()

        for b in b_array:
            reg2.insert(b)

        result = reg1 - reg2
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(result)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.elements.element import Element
from kqcircuits.elements.waveguide_coplanar_straight import WaveguideCoplanarStraight
from kqcircuits.elements.waveguide_coplanar_curved import WaveguideCoplanarCurved


class WaveguideCoplanar(Element):
    """The PCell declaration for an arbitrary coplanar waveguide.

    Coplanar waveguide defined by the width of the center conductor and gap. It can follow any segmented lines with
    predefined bending radios. It actually consists of straight and bent PCells. Termination lengths are lengths of
    extra ground gaps for opened transmission lines

    The ``path`` parameter defines the waypoints of the waveguide. When a DPath is supplied, the waypoints can be edited
    in the KLayout GUI with the Partial tool. Alternatively, a list of DPoint can be supplied, in which case the
    guiding shape is not visible in the GUI. This is useful for code-generated (sub)cells where graphical editing is not
    possible or desired.

    Warning:
        Arbitrary angle bents can have very small gaps between bends and straight segments due to
        precision of arithmetic. Small positive value of corner_safety_overlap can avoid these gaps.

    .. MARKERS_FOR_PNG 20,-10 50,0
    """

    path = Param(pdt.TypeShape, "TLine", pya.DPath([pya.DPoint(0, 0), pya.DPoint(100, 0)], 0))
    term1 = Param(pdt.TypeDouble, "Termination length start", 0, unit="μm")
    term2 = Param(pdt.TypeDouble, "Termination length end", 0, unit="μm")
    corner_safety_overlap = Param(pdt.TypeDouble, "Extend straight sections near corners", 0.001, unit="μm",
        docstring="Extend straight sections near corners by this amount (μm) to ensure all sections overlap")

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def parameters_from_shape_impl(self):
        points = [pya.DPoint(point * self.layout.dbu) for point in self.shape.each_point()]
        self.path = pya.DPath(points, 1)

    def transformation_from_shape_impl(self):
        return pya.Trans()

    def produce_waveguide(self):
        if isinstance(self.path, list):
            points = [pya.DPoint(p) if isinstance(p, pya.DVector) else p for p in self.path]
        else:
            points = list(self.path.each_point())

        if len(points) < 2:
            self.raise_error_on_cell("Need at least 2 points for a waveguide.",
                                     points[0] if len(points) == 1 else pya.DPoint())

        # distance between points[0] and beginning of the straight
        last_cut_dist = 0.0 if self.term1 == 0 else -self.corner_safety_overlap

        # For each segment except the last
        for i in range(0, len(points) - 2):
            # Check if straight can fit between points[i] and points[i + 1]
            v1, _, alpha1, alpha2, corner_pos = self.get_corner_data(points[i], points[i + 1], points[i + 2], self.r)
            alpha = (alpha2 - alpha1 + math.pi) % (2 * math.pi) - math.pi  # turn angle (between -pi and pi) in radians
            # distance between points[i + 1] and beginning of the straight
            cut_dist = self.r * math.tan(abs(alpha) / 2) - self.corner_safety_overlap
            straight_length = v1.length() - last_cut_dist - cut_dist
            if straight_length < 0:
                self.raise_error_on_cell("Straight segment cannot fit. Try decreasing the turn radius.",
                                         points[i] + v1 / 2)

            # Straight segment before corner
            if straight_length > self.corner_safety_overlap:
                cell_straight = self.add_element(WaveguideCoplanarStraight, l=straight_length)
                start_point = points[i] + last_cut_dist / v1.length() * v1
                transf = pya.DCplxTrans(1, math.degrees(alpha1), False, start_point)
                self.insert_cell(cell_straight, transf)

            # Curved segment at the corner
            if 2 * cut_dist >= self.corner_safety_overlap:
                cell_curved = self.add_element(WaveguideCoplanarCurved, alpha=alpha)
                transf = pya.DCplxTrans(1, math.degrees(alpha1) + (90 if alpha < 0 else -90), False, corner_pos)
                self.insert_cell(cell_curved, transf)

            # Prepare for next iteration
            last_cut_dist = cut_dist

        # Check if straight can fit between the last two points
        v1 = points[-1] - points[-2]
        cut_dist = 0.0 if self.term2 == 0 else -self.corner_safety_overlap
        straight_length = v1.length() - last_cut_dist - cut_dist
        if straight_length < 0:
            self.raise_error_on_cell("Straight segment cannot fit. Try decreasing the turn radius.",
                                     points[-2] + v1 / 2)

        # Straight segment at the end
        if straight_length > self.corner_safety_overlap:
            subcell = self.add_element(WaveguideCoplanarStraight, l=straight_length)
            start_point = points[-2] + last_cut_dist / v1.length() * v1
            transf = pya.DCplxTrans(1, math.degrees(math.atan2(v1.y, v1.x)), False, start_point)
            self.insert_cell(subcell, transf)

        # Termination before the first segment
        WaveguideCoplanar.produce_end_termination(self, points[1], points[0], self.term1)
        self.add_port("a", points[0], points[0] - points[1])

        # Terminate the end
        WaveguideCoplanar.produce_end_termination(self, points[-2], points[-1], self.term2)
        self.add_port("b", points[-1], points[-1] - points[-2])

    def build(self):
        self.produce_waveguide()

    @staticmethod
    def get_corner_data(point1, point2, point3, r):
        """Returns data needed to create a curved waveguide at path corner.

        Args:
            point1: point before corner
            point2: corner point
            point3: point after corner
            r: curve radius

        Returns:
            A tuple (``v1``, ``v2``, ``alpha1``, ``alpha2``, ``corner_pos``), where

            * ``v1``: the vector (`point2` - `point1`)
            * ``v2``: the vector (`point3` - `point2`)
            * ``alpha1``: angle between `v1` and positive x-axis
            * ``alpha2``: angle between `v2` and positive x-axis
            * ``corner_pos``: position where the curved waveguide should be placed

        """
        v1 = point2 - point1
        v2 = point3 - point2
        alpha1 = math.atan2(v1.y, v1.x)
        alpha2 = math.atan2(v2.y, v2.x)
        alpha = (alpha2 - alpha1 + math.pi) % (2 * math.pi) - math.pi  # turn angle (between -pi and pi) in radians
        alphacorner = alpha1 + (alpha + math.pi) / 2  # corner middle angle plus 90 degrees
        distcorner = (r if alpha > 0 else -r) / math.cos(alpha / 2)
        corner_pos = point2 + pya.DVector(math.cos(alphacorner)*distcorner, math.sin(alphacorner)*distcorner)
        return v1, v2, alpha1, alpha2, corner_pos

    @staticmethod
    def produce_end_termination(elem, point_1, point_2, term_len, face_index=0, opp_face_index=1):
        """Produces termination for a waveguide.

        The termination consists of a rectangular polygon in the metal gap layer, and grid avoidance around it.
        One edge of the polygon is centered at point_2, and the polygon extends to length "term_len" in the
        direction of (point_2 - point_1).

        Args:
            elem: Element from which the waveguide parameters for the termination are taken
            point_1: DPoint before point_2, used only to determine the direction
            point_2: DPoint after which termination is produced
            term_len (double): termination length, assumed positive
            face_index (int): face index of the face in elem where the termination is created
            opp_face_index (int): face index of the opposite face
        """
        a = elem.a
        b = elem.b

        v = (point_2 - point_1)*(1/point_1.distance(point_2))
        u = pya.DTrans.R270.trans(v)
        shift_start = pya.DTrans(pya.DVector(point_2))

        if term_len > 0:
            poly = pya.DPolygon([u*(a/2 + b), u*(a/2 + b) + v*term_len, u*(-a/2 - b) + v*term_len,
                                 u*(-a/2 - b)])
            elem.cell.shapes(elem.layout.layer(elem.face(face_index)["base_metal_gap_wo_grid"])).insert(
                poly.transform(shift_start))

        # protection
        term_len += elem.margin
        poly2 = pya.DPolygon([u*(a/2 + b + elem.margin), u*(a/2 + b + elem.margin) + v*term_len,
                              u*(-a/2 - b - elem.margin) + v*term_len, u*(-a/2 - b - elem.margin)])
        elem.add_protection(poly2.transform(shift_start), face_index, opp_face_index)

    @staticmethod
    def is_continuous(waveguide_cell, annotation_layer, tolerance):
        """Returns true if the given waveguide is determined to be continuous, false otherwise.

        The waveguide is considered continuous if the endpoints of its every segment (except first and last) are close
        enough to the endpoints of neighboring segments. The waveguide segments are not necessarily ordered correctly
        when iterating through the cells using begin_shapes_rec. This means we must compare the endpoints of each
        waveguide segment to the endpoints of all other waveguide segments.

        Args:
            waveguide_cell: Cell of the waveguide.
            annotation_layer: unsigned int representing the annotation layer
            tolerance: maximum allowed distance between connected waveguide segments

        """
        is_continuous = True

        # find the two endpoints for every waveguide segment

        endpoints = []  # endpoints of waveguide segment i are contained in endpoints[i][0] and endpoints[i][1]
        shapes_iter = waveguide_cell.begin_shapes_rec(annotation_layer)

        while not shapes_iter.at_end():
            shape = shapes_iter.shape()
            if shape.is_path():
                dtrans = shapes_iter.dtrans()  # transformation from shape coordinates to waveguide_cell coordinates
                pts = shape.each_dpoint()
                first_point = dtrans * next(pts, None)
                last_point = first_point.dup()
                for pt in pts:
                    last_point = pt
                last_point = dtrans * last_point
                endpoints.append([first_point, last_point])
            shapes_iter.next()

        # for every waveguide segment endpoint, try to find another endpoint which is close to it

        num_segments = len(endpoints)
        num_non_connected_points = 0

        for i in range(num_segments):

            def find_connected_point(point):
                """Tries to find a waveguide segment endpoint close enough to the given point."""

                found_connected_point = False

                for j in range(num_segments):
                    # pylint: disable=cell-var-from-loop
                    if i != j and (point.distance(endpoints[j][1]) < tolerance
                                   or point.distance(endpoints[j][0]) < tolerance):
                        # print("{} | {} | {}".format(point, endpoints[j][1], endpoints[j][0]))
                        found_connected_point = True
                        break

                if not found_connected_point:
                    nonlocal num_non_connected_points
                    num_non_connected_points += 1

            if endpoints[i][0].distance(endpoints[i][1]) != 0:  # we ignore any zero-length segments

                find_connected_point(endpoints[i][0])
                find_connected_point(endpoints[i][1])

            # we can have up to 2 non-connected points, because ends of the waveguide don't have to be connected
            if num_non_connected_points > 2:
                is_continuous = False
                break

        return is_continuous



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.fluxlines.fluxline import Fluxline


@logged
@add_parameters_from(Fluxline, fluxline_gap_width=3)
class FluxlineStraight(Fluxline):
    """Fluxline variant "straight vertical".

     .. MARKERS_FOR_PNG -4,-6 0,-7
    """
    fluxline_width = Param(pdt.TypeDouble, "Fluxline width", 0, unit="μm", hidden=True)

    def build(self):

        b = self.fluxline_gap_width
        a = (self.a/self.b)*b  # fluxline center width
        l1 = 2*b  # straight down
        bottom_y = -b - l1

        # origin at edge of the qubit gap

        # Right gap of the fluxline. Points created clockwise starting from top left point.
        x_offset = 0  # a + b
        right_gap_pts = [
            pya.DPoint(-b / 2 + x_offset, -b),
            pya.DPoint(b / 2 + x_offset, -b),
            pya.DPoint(b / 2 + x_offset, bottom_y),
            pya.DPoint(-b / 2 + x_offset, bottom_y),
        ]
        right_gap = pya.DPolygon(right_gap_pts)

        # Left gap of the fluxline. Points created clockwise starting from top left point.
        left_gap_pts = [
            right_gap_pts[0] + pya.DPoint(-a - b, 0),
            right_gap_pts[1] + pya.DPoint(-a - b, 0),
            right_gap_pts[2] + pya.DPoint(-a - b, 0),
            right_gap_pts[3] + pya.DPoint(-a - b, 0),
        ]
        left_gap = pya.DPolygon(left_gap_pts)

        self._insert_fluxline_shapes(left_gap, right_gap)
        self._add_fluxline_refpoints(pya.DPoint(-a/2 - b/2 + x_offset, bottom_y))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).



class LayerCluster:
    """Container class for a cluster of layers.

    Some files need to be exported with certain clusters of layers, for example for EBL process. This class wraps
    information about such layer cluster.

    Attributes:
        main_layers: list of names of the layers which should contain shapes for this cluster to be exported
        extra_layers: list of names of other layers exported together with main_layers
        face_id: face_id of the layers in this cluster

    """

    def __init__(self, main_layers, extra_layers, face_id):
        self.main_layers = main_layers
        self.extra_layers = extra_layers
        self.face_id = face_id

    def all_layers(self):
        """Returns main_layers + extra_layers."""
        return self.main_layers + self.extra_layers



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.test_structures.airbridge_dc import AirbridgeDC


class AirbridgeDcTest(Chip):
    """Chip full of airbridge 4-point dc tests."""

    n_step = Param(pdt.TypeInt, "Increment step for number of airbridges", 1)
    test_width = Param(pdt.TypeDouble, "Width of a single test", 2000, unit="μm")

    def build(self):

        d1 = float(self.frames_dice_width[0]) + self.dice_grid_margin # smaller distance of test area from chip edge
        d2 = 2000  # larger distance of test area from chip edge
        chip_size = self.box.width()

        n_ab = self.n_step
        test_id = 0

        n_ab, test_id = self._produce_tests_within_box(pya.DBox(d2, chip_size - d2, chip_size - d2, chip_size - d1),
                                                       n_ab, test_id)
        n_ab, test_id = self._produce_tests_within_box(pya.DBox(d1, d2, chip_size - d1, chip_size - d2), n_ab, test_id)
        n_ab, test_id = self._produce_tests_within_box(pya.DBox(d2, d1, chip_size - d2, d2), n_ab, test_id)

    def _produce_tests_within_box(self, box, n_ab, test_id):

        num_horizontal = int(box.width()//self.test_width)
        x_start = box.left + (box.width() - self.test_width*num_horizontal)/2  # to make them horizontally centered
        x = x_start
        y = box.top

        while y > box.bottom:

            while x < box.right:

                cell = self.add_element(AirbridgeDC, n_ab=n_ab, width=self.test_width)
                test_height = cell.dbbox_per_layer(self.get_layer("base_metal_gap_wo_grid")).height()
                num_vertical = int(box.height()//test_height)  # assuming test_height same for all tests
                y_offset = (box.height() - test_height*num_vertical)/2  # to make them vertically centered

                trans = pya.DTrans(pya.DVector(x + self.test_width/2, y - y_offset - test_height/2))

                x += self.test_width
                if x > box.right or y - test_height < box.bottom:
                    break

                label_trans = pya.DCplxTrans(0.8, 0, False, -self.test_width/2, 0)
                self.insert_cell(cell, trans, "test_{}".format(test_id), label_trans)

                n_ab += self.n_step
                test_id += 1

            x = x_start
            y -= test_height

        return n_ab, test_id



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import os.path
from pathlib import Path

from kqcircuits.pya_resolver import pya
from kqcircuits.defaults import default_layers
from kqcircuits.simulations.export.sonnet import parser
from kqcircuits.simulations.export.util import find_edge_from_point_in_polygons
from kqcircuits.simulations.port import InternalPort, EdgePort
from kqcircuits.simulations.simulation import Simulation, get_simulation_layer_by_name
from kqcircuits.util.export_helper import write_commit_reference_file


def poly_and_edge_indices(polygons, dbu, port, number, location, group):
    i, j, _ = find_edge_from_point_in_polygons(
        polygons,
        location,
        dbu,
        tolerance=5.0  # hardcoded, feel free to change
        )

    return parser.port(
        portnum=number,
        ipolygon=i + 1,
        ivertex=j,
        port_type=("CUP" if group else "STD"),
        group=group,
        resist=port.resistance,
        react=port.reactance,
        induct=port.inductance,
        capac=port.capacitance
    )


def export_sonnet_son(simulation: Simulation, path: Path, detailed_resonance=False, lower_accuracy=False, current=False,
                      control='ABS', fill_type='Staircase', simulation_safety=0):
    """
    Export simulation into son file.

    Arguments:
        simulation: The simulation to be exported.
        path: Location where to write son files.
        detailed_resonance: More info in
            www.sonnetsoftware.com/support/downloads/techdocs/Enhanced_Resonance_Detection_Feature.pdf
        lower_accuracy: False sets Sonnet to Fine/Edge meshing and True to Coarse/Edge Meshing.
        fill_type: Value 'Staircase' sets the default fill type for polygons. The other option is 'Conformal' which can
            be faster but less accurate. A good workflow could be to set everything to Staircase and then set some
            meanders to Conformal in Sonnet.
        current: True computes currents in Sonnet which adds to simulation time but can used to easily see connection
            errors.
        control: Selects what analysis control is used in Sonnet. Options are 'Simple', 'ABS' and 'Sweep' for parameter
            sweeping.
        simulation_safety: Adds extra ground area to a simulation environment (in µm).

    Returns:
        Path to exported son file.
    """
    if simulation is None or not isinstance(simulation, Simulation):
        raise ValueError("Cannot export without simulation")

    def get_sonnet_strings(material_type, grid_size, symmetry):
        layout = simulation.cell.layout()
        dbu = layout.dbu
        layer_pad = layout.layer(get_simulation_layer_by_name("1t1_airbridge_pads"))
        layer_bridge = layout.layer(get_simulation_layer_by_name("1t1_airbridge_flyover"))
        layer_son = layout.layer(get_simulation_layer_by_name("1t1_signal"))
        layer_son_ground = layout.layer(get_simulation_layer_by_name("1t1_ground"))

        simpolygons = [p.polygon for p in simulation.cell.shapes(layer_son).each()] + \
                      [p.polygon for p in simulation.cell.shapes(layer_son_ground).each()]
        airbridge_polygons = [p.polygon for p in simulation.cell.shapes(layer_bridge).each()]
        airpads_polygons = [p.polygon for p in simulation.cell.shapes(layer_pad).each()]
        for p in airpads_polygons:
            p.isVia = True

        level_iter = iter(len(simpolygons) * [(2 if material_type == "Si+Al" else 0)] +
                          len(airbridge_polygons) * [1] + len(airpads_polygons) * [2])

        polys = parser.polygons(simpolygons + airbridge_polygons + airpads_polygons,
                                pya.DVector(-simulation.box.p1.x, -simulation.box.p2.y), dbu,
                                # get the bottom left corner
                                ilevel=level_iter, fill_type=("V" if (fill_type == "Conformal") else "N")
                                )

        # find port edges
        sstring_ports = ""
        refplane_dirs = []
        port_ipolys = []
        group_ascii = ord('A')
        calgroup = ''

        if simulation.use_ports:
            for port in simulation.ports:
                if isinstance(port, InternalPort):
                    sstring_ports += poly_and_edge_indices(
                        simpolygons + airbridge_polygons, dbu,
                        port, port.number, port.signal_location, chr(group_ascii))
                    sstring_ports += poly_and_edge_indices(
                        simpolygons + airbridge_polygons, dbu,
                        port, -port.number, port.ground_location, chr(group_ascii))
                    calgroup += 'CUPGRP "{}"\nID 28\nGNDREF F\nTWTYPE FEED\nEND\n'.format(chr(group_ascii))
                    group_ascii += 1
                elif isinstance(port, EdgePort):
                    # TODO: re-implement calibration
                    #
                    #     refplane_dirs.append(port.side)
                    #     ipoly = poly_and_edge_indeces(cell, simpolygons + airbridge_polygons, dbu, port, ls)
                    #     logging.info(re.findall(r'POLY (\d+)', ipoly))
                    #     port_ipolys.append(re.findall(r'POLY (\d+)', ipoly)) # scan ipolygon
                    sstring_ports += poly_and_edge_indices(
                        simpolygons + airbridge_polygons, dbu,
                        port, port.number, port.signal_location, "")

        sonnet_box = parser.box(
            xwidth=simulation.box.width(),
            ywidth=simulation.box.height(),
            xcells=int(simulation.box.width() / grid_size),
            ycells=int(simulation.box.height() / grid_size),
            materials_type=material_type
        )

        return {
            "polygons": polys,
            "box": sonnet_box,
            "ports": sstring_ports,
            "calgroup": calgroup,
            "refplanes": parser.refplanes(refplane_dirs, simulation_safety, port_ipolys),
            "symmetry": parser.symmetry(symmetry),
            "nports": len({abs(port.number) for port in simulation.ports}),
            "resonance_abs": "DET_ABS_RES Y" if detailed_resonance else "DET_ABS_RES N",
            "lower_accuracy": "1" if lower_accuracy else "0",
            "current": "j" if current else ""
        }

    # detect airbridges
    shapes_in_air = simulation.layout.begin_shapes(simulation.cell, simulation.layout.layer(
        default_layers["1t1_airbridge_flyover"]))
    materials_type = "Si+Al" if not shapes_in_air.shape().is_null() else "Si BT"

    sonnet_strings = get_sonnet_strings(materials_type, 1, False)
    sonnet_strings["control"] = parser.control(control)

    son_filename = str(path.joinpath(simulation.name + '.son'))
    parser.apply_template(
        os.path.join(os.path.dirname(os.path.abspath(parser.__file__)), "template.son"),
        son_filename,
        sonnet_strings
    )
    return son_filename


def export_sonnet(simulations, path: Path, detailed_resonance=False, lower_accuracy=False, current=False, control='ABS',
                  fill_type='Staircase', simulation_safety=0):
    """
    Export Sonnet simulations by writing son files.

    Arguments:
        simulations: List of simulations to be exported.
        path: Location where to write export files.
        detailed_resonance: More info in
            www.sonnetsoftware.com/support/downloads/techdocs/Enhanced_Resonance_Detection_Feature.pdf
        lower_accuracy: False sets Sonnet to Fine/Edge meshing and True to Coarse/Edge Meshing.
        fill_type: Value 'Staircase' sets the default fill type for polygons. The other option is 'Conformal' which can
            be faster but less accurate. A good workflow could be to set everything to Staircase and then set some
            meanders to Conformal in Sonnet.
        current: True computes currents in Sonnet which adds to simulation time but can used to easily see connection
            errors.
        control: Selects what analysis control is used in Sonnet. Options are 'Simple', 'ABS' and 'Sweep' for parameter
            sweeping.
        simulation_safety: Adds extra ground area to a simulation environment (in µm).

    Returns:
        List of paths to exported son files.
    """
    write_commit_reference_file(path)
    son_filenames = []
    for simulation in simulations:
        son_filenames.append(export_sonnet_son(simulation, path, detailed_resonance=detailed_resonance,
                                               lower_accuracy=lower_accuracy, current=current, control=control,
                                               fill_type=fill_type, simulation_safety=simulation_safety))
    return son_filenames



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from math import cos, sin, pi, radians

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.geometry_helper import arc_points
from kqcircuits.elements.element import Element
from kqcircuits.elements.airbridges.airbridge import Airbridge


@add_parameters_from(Airbridge, "airbridge_type")
class WaveguideCoplanarSplitter(Element):
    """
    The PCell declaration of a multiway waveguide splitter. The number of ports is defined by the length of the
    parameter lists. Ports are labelled by letters starting from ``a``

    .. MARKERS_FOR_PNG -9,4 -1,-8 9,9
    """
    lengths = Param(pdt.TypeList, "Waveguide length per port, measured from origin", [11, 11, 11])
    angles = Param(pdt.TypeList, "Angle of each port (degrees)", [0, 180, 270])
    use_airbridges = Param(pdt.TypeBoolean, "Use airbridges at a distance from the centre", False)
    bridge_distance = Param(pdt.TypeDouble, "Bridges distance from centre", 80)
    a_list = Param(pdt.TypeList, "Center conductor widths", [], unit="[μm]",
                   docstring="List of center conductor widths for each port."
                             " If empty, self.a will be used for all ports instead. [μm]")
    b_list = Param(pdt.TypeList, "Gap widths", [], unit="[μm]",
                   docstring="List of gap widths for each port."
                             " If empty, self.b will be used for all ports instead. [μm]")
    port_names = Param(pdt.TypeList, "Port names", ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j'])

    def build(self):

        gap_shapes = []
        trace_shapes = []
        avoidance_shapes = []

        # Tolerance to make sure that the trace shape is larger than the gap shape after integer conversion
        rounding_tolerance = 10 * self.layout.dbu

        # Convert a, b to lists of right length
        a_list = self.a_list if (len(self.a_list) > 0 and self.a_list[0] != "") else [self.a] * len(self.angles)
        b_list = self.b_list if (len(self.b_list) > 0 and self.b_list[0] != "") else [self.b] * len(self.angles)

        for length_str, angle_str, port_name, a, b in zip(self.lengths, self.angles, self.port_names, a_list, b_list):
            angle_deg = float(angle_str)
            angle_rad = radians(angle_deg)
            length = float(length_str)
            a = float(a)
            b = float(b)

            # Generate port shapes
            gap_shapes.append(self._get_port_shape(
                angle_rad=angle_rad,
                length=length,
                width=a + 2*b
            ).to_itype(self.layout.dbu))

            trace_shapes.append(self._get_port_shape(
                angle_rad=angle_rad,
                length=length + rounding_tolerance,
                width=a
            ).to_itype(self.layout.dbu))

            avoidance_shapes.append(self._get_port_shape(
                angle_rad=angle_rad,
                length=length + self.margin,
                width=a + 2*b + 2*self.margin
            ).to_itype(self.layout.dbu))

            # Port refpoints
            self.add_port(
                port_name,
                pya.DPoint(length*cos(angle_rad), length*sin(angle_rad)),
                pya.DVector(self.r*cos(angle_rad), self.r*sin(angle_rad))
            )

            # Waveguide length annotation
            self.cell.shapes(self.get_layer("waveguide_path")).insert(
                pya.DPath([self.refpoints[f"port_{port_name}"], self.refpoints["base"]], a)
            )

            # Airbridges
            if self.use_airbridges:
                ab_trans = pya.DCplxTrans(1, angle_deg, False,
                                          self.bridge_distance*cos(angle_rad),
                                          self.bridge_distance*sin(angle_rad))
                ab_cell = self.add_element(Airbridge, pad_length=14, pad_extra=2)
                self.insert_cell(ab_cell, ab_trans)

        # Merge and insert shapes
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(
            pya.Region(gap_shapes) - pya.Region(trace_shapes)
        )
        self.add_protection(pya.Region(avoidance_shapes).merged())

    def _get_port_shape(self, angle_rad, length, width):
        # Generate a shape consisting of a rectangle (length, width) starting at (0, 0), with a round cap at the origin
        # side.

        r = width/2  # Radius of round cap

        # Straight section
        c = cos(angle_rad)
        s = sin(angle_rad)
        points = [pya.DPoint(length * c - r * s, length * s + r * c),
                  pya.DPoint(length * c + r * s, length * s - r * c)]

        # Corner section
        angles = [radians(float(angle)) for angle in self.angles]
        prev_rad = min([2 * pi if a == angle_rad else (angle_rad - a) % (2 * pi) for a in angles])
        next_rad = min([2 * pi if a == angle_rad else (a - angle_rad) % (2 * pi) for a in angles])
        if prev_rad <= pi / 2:
            dist = r * cos(prev_rad) / sin(prev_rad)
            if length > dist:
                points.append(pya.DPoint(dist * c + r * s, dist * s - r * c))
        else:
            points += arc_points(r, angle_rad - pi / 2, angle_rad - prev_rad, self.n)
        points.append(pya.DPoint(0.0, 0.0))
        if next_rad <= pi / 2:
            dist = r * cos(next_rad) / sin(next_rad)
            if length > dist:
                points.append(pya.DPoint(dist * c - r * s, dist * s + r * c))
        else:
            points += arc_points(r, angle_rad + next_rad, angle_rad + pi / 2, self.n)

        return pya.DPolygon(points)


def t_cross_parameters(a=Element.get_schema()["a"].default, b=Element.get_schema()["b"].default,
                      a2=Element.a, b2=Element.b, length_extra=0, length_extra_side=0, **kwargs):
    """A utility function to easily produce T-cross splitter (old WaveguideCoplanarTCross).

    Args:
        a: Width of center conductor
        b: Width of gap
        a2: Center conductor width of the side waveguide
        b2: Gap of the side waveguide
        length_extra: Extra length
        length_extra_side: Extra length of the side waveguide

    Returns:
        dictionary of parameters for WaveguideCoplanarSplitter
    """
    length = a2 / 2 + b2 + length_extra
    length2 = a / 2 + b + length_extra_side
    return {'lengths': [length, length, length2],
            'angles': [0, 180, 270],
            'a_list': [a, a, a2],
            'b_list': [b, b, b2],
            'port_names': ['right', 'left', 'bottom'],
            **kwargs}



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.meander import Meander
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.port import InternalPort
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.util.parameters import Param, pdt


class XMonsDirectCouplingFullChipSim(Simulation):

    qubit_spacing = Param(pdt.TypeDouble, "Qubit spacing", 10, unit="μm")
    arm_width_a = Param(pdt.TypeDouble, "Qubit A and C arm width", 24, unit="μm")
    arm_width_b = Param(pdt.TypeDouble, "Qubit B arm width", 24, unit="μm")
    enable_flux_lines = Param(pdt.TypeBoolean, "To flux or not to flux", True)
    enable_drive_lines = Param(pdt.TypeBoolean, "To drive or not to drive", True)
    enable_transmission_line = Param(pdt.TypeBoolean, "To transmit?", True)


    def produce_waveguide(self, path, term1=0, term2=0, turn_radius=None):
        if turn_radius is None:
            turn_radius = self.r

        tl = self.add_element(WaveguideCoplanar,
            path=pya.DPath(path, 1),
            r=turn_radius,
            term1=term1,
            term2=term2,
        )
        self.cell.insert(pya.DCellInstArray(tl.cell_index(), pya.DTrans()))

        return tl.length()

    def produce_qubit(self, qubit_cell, center_x, center_y=5e3, name=None):
        qubit_trans = pya.DTrans(0, False, center_x, center_y)
        qubit_inst = self.cell.insert(pya.DCellInstArray(qubit_cell.cell_index(), qubit_trans))
        if name:
            qubit_inst.set_property("id", name)

        refpoints_abs = self.get_refpoints(qubit_cell, qubit_inst.dtrans)
        port_qubit_dr = refpoints_abs["port_drive"]
        port_qubit_fl = refpoints_abs["port_flux"] if "port_flux" in refpoints_abs else None
        port_qubit_ro = refpoints_abs["port_cplr1"]
        port_qubit_squid_a = refpoints_abs["port_squid_a"]
        port_qubit_squid_b = refpoints_abs["port_squid_b"]

        return (port_qubit_dr, port_qubit_fl, port_qubit_ro, port_qubit_squid_a, port_qubit_squid_b)

    def produce_readout_resonator(self, pos_start, end_y, length):
        width_rr = 300

        # coupler
        pos_coupler_end = pya.DPoint(pos_start.x, end_y - 3 * self.r)
        len_coupler = self.produce_waveguide([
            pya.DPoint(pos_start.x - width_rr / 2, end_y),
            pya.DPoint(pos_start.x + width_rr / 2, end_y),
            pya.DPoint(pos_start.x + width_rr / 2, end_y - 2 * self.r),
            pya.DPoint(pos_start.x, end_y - 2 * self.r),
            pos_coupler_end
        ], turn_radius=50, term1=10)

        # meander
        meander = self.add_element(Meander,
            start=pos_coupler_end,
            end=pos_start,
            length=length - len_coupler,
            meanders=8,
            r=50
        )
        self.cell.insert(pya.DCellInstArray(meander.cell_index(), pya.DTrans()))

    def produce_launcher(self, pos, direction):
        """Wrapper function for launcher PCell placement at `pos` with `direction`, `name` and `width`."""

        subcell = self.add_element(WaveguideCoplanar,
            path= pya.DPath([pya.DPoint(0, 0), pya.DPoint(90, 0)], 0),
            term2=10,
        )
        subcell2 = self.add_element(WaveguideCoplanar,
            path=pya.DPath([pya.DPoint(100, 0), pya.DPoint(110, 0)], 0),
            term2=0,
        )

        if isinstance(direction, str):
            direction = {"E": 0, "W": 180, "S": -90, "N": 90}[direction]
        transf = pya.DCplxTrans(1, direction, False, pos)
        self.cell.insert(pya.DCellInstArray(subcell.cell_index(), transf))
        self.cell.insert(pya.DCellInstArray(subcell2.cell_index(), transf))

    def produce_launchers_SMA8(self, enabled=["WS", "WN", "ES", "EN", "SW", "SE", "NW", "NE"]):
        """Produces enabled launchers for SMA8 sample holder default locations

        Args:
            enabled: List of enabled standard launchers from set ("WS", "WN", "ES", "EN", "SW", "SE", "NW", "NE")

        Effect:
            launchers PCells added to the class parent cell.

        Returns:
            launchers dictionary, where keys are launcher names and values are tuples of (point, heading, distance from
            chip edge)
        """
        # pylint: disable=invalid-name,dangerous-default-value
        # dictionary of point, heading, distance from chip edge
        launchers = {
            "WS": (pya.DPoint(800, 2800), "W", 300),
            "ES": (pya.DPoint(9200, 2800), "E", 300),
            "WN": (pya.DPoint(800, 7200), "W", 300),
            "EN": (pya.DPoint(9200, 7200), "E", 300),
            "SW": (pya.DPoint(2800, 800), "S", 300),
            "NW": (pya.DPoint(2800, 9200), "N", 300),
            "SE": (pya.DPoint(7200, 800), "S", 300),
            "NE": (pya.DPoint(7200, 9200), "N", 300)
        }
        for name in enabled:
            self.produce_launcher(launchers[name][0], launchers[name][1])
        return launchers

    def build(self):
        enabled_launchers = []
        launchers_with_ports = []
        if self.enable_transmission_line:
            enabled_launchers += ['NW', 'NE']
            launchers_with_ports += ['NW', 'NE']
        if self.enable_drive_lines:
            enabled_launchers += ['WN', 'SW', 'ES']
            launchers_with_ports += ['WN', 'SW', 'ES']
        if self.enable_flux_lines:
            enabled_launchers += ['WS', 'SE', 'EN']
            # For now, flux lines don't have ports
            # TODO: Set up a way to make flux line a different polygon from ground plane, and move it to signal layer
        launchers = self.produce_launchers_SMA8(enabled=enabled_launchers)

        # Finnmon
        qubit_props_common = {
            "fluxline_type": "Fluxline Standard" if self.enable_flux_lines else "none",
            "junction_type": 'Sim',
            "arm_length": [146] * 4,
            "island_r": 2,
            "cpl_length": [0, 140, 0],
            "cpl_width": [60, 24, 60],
            "cpl_gap": [110, 102, 110],
            "cl_offset": [150, 150]
        }
        finnmon_a = self.add_element(Swissmon,
            arm_width=[self.arm_width_a] * 4,
            gap_width=[(72 - self.arm_width_a) / 2] * 4, **qubit_props_common)
        finnmon_b = self.add_element(Swissmon,
            arm_width=[self.arm_width_b] * 4,
            gap_width=[(72 - self.arm_width_b) / 2] * 4, **qubit_props_common)

        (pos_qb1_dr, pos_qb1_fl, pos_qb1_rr, port_qubit1_squid_a, port_qubit1_squid_b) = \
            self.produce_qubit(finnmon_a, 5e3 - 330 - self.qubit_spacing, name="qb_1")
        (pos_qb2_dr, pos_qb2_fl, pos_qb2_rr, port_qubit2_squid_a, port_qubit2_squid_b) = \
            self.produce_qubit(finnmon_b, 5e3, name="qb_`2")
        (pos_qb3_dr, pos_qb3_fl, pos_qb3_rr, port_qubit3_squid_a, port_qubit3_squid_b) = \
            self.produce_qubit(finnmon_a, 5e3 + 330 + self.qubit_spacing, name="qb_3")

        # Readout resonators
        height_rr_feedline = 7.3e3
        self.produce_readout_resonator(pos_qb1_rr, height_rr_feedline - 30, 4330.9)  # values from manual X06
        self.produce_readout_resonator(pos_qb2_rr, height_rr_feedline - 30, 4225.9)
        self.produce_readout_resonator(pos_qb3_rr, height_rr_feedline - 30, 4177.9)

        # Transmission lines
        tl_gap = 300

        if self.enable_transmission_line:
            # RR feedline
            self.produce_waveguide([
                launchers["NW"][0],
                pya.DPoint(launchers["NW"][0].x, height_rr_feedline),
                pya.DPoint(launchers["NE"][0].x, height_rr_feedline),
                launchers["NE"][0]])

        if self.enable_drive_lines:
            # Qb1 chargeline
            self.produce_waveguide([
                launchers["WN"][0],
                pya.DPoint(launchers["NW"][0].x - tl_gap, launchers["WN"][0].y),
                pya.DPoint(launchers["NW"][0].x - tl_gap, launchers["WS"][0].y + tl_gap),
                pya.DPoint(pos_qb1_dr.x, launchers["WS"][0].y + tl_gap),
                pos_qb1_dr], term2=self.b)
            # Qb2 chargeline
            self.produce_waveguide([
                launchers["SW"][0],
                pya.DPoint(launchers["SW"][0].x, launchers["WS"][0].y - tl_gap),
                pya.DPoint(pos_qb2_dr.x, launchers["WS"][0].y - tl_gap),
                pos_qb2_dr], term2=self.b)
            # Qb3 driveline
            self.produce_waveguide([
                launchers["ES"][0],
                pya.DPoint(pos_qb3_dr.x, launchers["ES"][0].y),
                pos_qb3_dr], term2=self.b)

        if self.enable_flux_lines:
            # Qb1 fluxline
            self.produce_waveguide([
                launchers["WS"][0],
                pya.DPoint(pos_qb1_fl.x, launchers["WS"][0].y),
                pos_qb1_fl])
            # Qb2 fluxline
            self.produce_waveguide([
                launchers["SE"][0],
                pya.DPoint(launchers["SE"][0].x, launchers["ES"][0].y - tl_gap),
                pya.DPoint(pos_qb2_fl.x, launchers["ES"][0].y - tl_gap),
                pos_qb2_fl])
            # Qb3 fluxline
            self.produce_waveguide([
                launchers["EN"][0],
                pya.DPoint(launchers["NE"][0].x + tl_gap, launchers["EN"][0].y),
                pya.DPoint(launchers["NE"][0].x + tl_gap, launchers["ES"][0].y + tl_gap),
                pya.DPoint(pos_qb3_fl.x, launchers["ES"][0].y + tl_gap),
                pos_qb3_fl])


        dx = {'W': -90, 'E': 90, 'S': 0, 'N': 0}
        dy = {'W': 0, 'E': 0, 'S': -90, 'N': 90}

        dx2 = {'W': -100, 'E': 100, 'S': 0, 'N': 0}
        dy2 = {'W': 0, 'E': 0, 'S': -100, 'N': 100}


        for i, launcher_name in enumerate(launchers_with_ports):
            launcher = launchers[launcher_name]
            p1 = pya.DPoint(launcher[0].x + dx[launcher[1]], launcher[0].y + dy[launcher[1]])
            p2 = pya.DPoint(launcher[0].x + dx2[launcher[1]], launcher[0].y + dy2[launcher[1]])
            self.ports.append(InternalPort(i + 1, *self.etched_line(p1, p2)))

        self.ports.append(InternalPort(9, *self.etched_line(port_qubit1_squid_a, port_qubit1_squid_b)))
        self.ports.append(InternalPort(10, *self.etched_line(port_qubit2_squid_a, port_qubit2_squid_b)))
        self.ports.append(InternalPort(11, *self.etched_line(port_qubit3_squid_a, port_qubit3_squid_b)))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from typing import List
from kqcircuits.pya_resolver import pya
from kqcircuits.defaults import default_output_format


def export_layers(filename, layout, cells=None, layers=None, output_format=default_output_format):

    svopt = pya.SaveLayoutOptions()
    svopt.format = output_format
    svopt.write_context_info = False
    if layers is not None:
        svopt.deselect_all_layers()
        for layer in layers:
            svopt.add_layer(layout.layer(layer), layer)
    if cells is not None:
        svopt.clear_cells()
        for cell in cells:
            svopt.add_cell(cell.cell_index())
    svopt.no_empty_cells = True

    layout.write(filename, svopt)


def find_edge_from_point_in_cell(cell: pya.Cell, layer: int, point: pya.DPoint, dbu, tolerance=0.01):
    """
    Finds the edge closest to a point, and returns the edge as well as it's polygon and edge index
    """
    return find_edge_from_point_in_polygons(cell.shapes(layer).each(pya.Shapes.SPolygons), point, dbu, tolerance)


def find_edge_from_point_in_polygons(polygons: List[pya.Polygon], point: pya.DPoint, dbu, tolerance=0.01):
    """
    Finds the edge closest to a point, and returns the edge as well as it's polygon and edge index
    """
    # Find closest edge to point
    edges = [(i, j, edge.to_dtype(dbu))
             for (i, polygon) in enumerate(polygons)
             for (j, edge) in enumerate(polygon.each_edge())
             ]
    (distance, i, j, nearest_edge) = \
        sorted([(edge.distance_abs(point), i, j, edge) for (i, j, edge) in edges])[0]
    if distance < tolerance:
        return i, j, nearest_edge
    else:
        raise ValueError(f"No edge found at {point=}, {nearest_edge=}, {distance=}")


def get_enclosing_polygon(points: List[List[float]]):
    """
    Order points in such a way that they form a polygon without intersecting
    lines. The ordering is clockwise starting from the left-most point.

    Arguments:
        points: List of points [x,y]

    Returns:
        ordered list of points [x,y]
    """

    # Find y-coordinate of linear interpolation between p0 = [x0,y0] and
    # p1 = [x1,y1] corresponding to x
    def _linearinterpy(p0, p1, x):
        """
        Find y-coordinate of linear interpolation between p0 = [x0,y0] and p1 = [x1,y1] corresponding to x

        Arguments:
            p0, p1: Points [x,y]
            x: x-coordinate to interpolate at

        Returns:
            y = y0 + (x-x0)*(dy/dx)
        """
        return p0[1] + (x-p0[0])*((p1[1]-p0[1])/(p1[0]-p0[0]))

    # Sort by x and then y, to ensure we go from lowest left-most point to
    # highest right-most point.
    points.sort()

    # Leftmost and rightmost point
    pleft = points[0]
    pright = points[-1]

    # Split remaining points into groups above and below
    # the line pleft - pright
    pabove = []
    pbelow = []
    for p in points[1:-1]:
        if p[1] > _linearinterpy(pleft, pright, p[0]):
            pabove.append(p)
        else:
            pbelow.append(p)

    # Construct polygon starting from pleft and going clockwise
    # Note: we rely on the fact that pabove and pbelow are still sorted by x
    # Note: the polygon is not closed.
    pbelow.reverse()
    return [pleft] + pabove + [pright] + pbelow



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt


class WaveguideCoplanarStraight(Element):
    """The PCell declaration of a straight segment of a coplanar waveguide.

   .. MARKERS_FOR_PNG 15,8 15,0
    """

    l = Param(pdt.TypeDouble, "Length", 30)

    def build(self):
        # Refpoint in the first end
        # Left gap
        pts = [
            pya.DPoint(0, self.a / 2 + 0),
            pya.DPoint(self.l, self.a / 2 + 0),
            pya.DPoint(self.l, self.a / 2 + self.b),
            pya.DPoint(0, self.a / 2 + self.b)
        ]
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        # Right gap
        pts = [
            pya.DPoint(0, -self.a / 2 + 0),
            pya.DPoint(self.l, -self.a / 2 + 0),
            pya.DPoint(self.l, -self.a / 2 - self.b),
            pya.DPoint(0, -self.a / 2 - self.b)
        ]
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        # Protection layer
        w = self.a / 2 + self.b + self.margin
        pts = [
            pya.DPoint(0, -w),
            pya.DPoint(self.l, -w),
            pya.DPoint(self.l, w),
            pya.DPoint(0, w)
        ]
        self.add_protection(pya.DPolygon(pts))
        # Waveguide length
        pts = [
            pya.DPoint(0, 0),
            pya.DPoint(self.l, 0),
        ]
        shape = pya.DPath(pts, self.a)
        self.cell.shapes(self.get_layer("waveguide_path")).insert(shape)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import os
from kqcircuits.defaults import node_editor_valid_elements
from kqcircuits.pya_resolver import pya
from kqcircuits.util.gui_helper import node_from_text, get_nodes_near_position, replace_node, node_to_text


class EditNodePlugin(pya.Plugin):
    def __init__(self, manager, view):
        self.is_active = False
        self.manager = manager
        self.view = view
        self.selection = None
        self.last_dialog_position = None
        self.capture_range = 10  # Mouse capture and marker size in pixels
        self.last_mouse_position = pya.DPoint(0, 0)
        self.create_dialog()

    def create_dialog(self):
        dialog = pya.QDialog(pya.Application.instance().main_window())
        self.dialog = dialog
        dialog.windowTitle = 'Edit Node'
        dialog.setModal(False)

        form = pya.QFormLayout(dialog)

        def add_row(label, widget):
            form.addRow(label, widget)
            return widget

        self.text_x = add_row("x position (µm): ", pya.QLineEdit("0.0", dialog))
        self.text_y = add_row("y position (µm):", pya.QLineEdit("0.0", dialog))

        form.addRow(pya.QLabel("Optional arguments, leave empty for default:", dialog))

        self.text_inst_name = add_row("Instance name:", pya.QLineEdit("", dialog))
        self.text_angle = add_row("Angle (degrees):", pya.QLineEdit("", dialog))
        self.text_length_before = add_row("Length before (µm):", pya.QLineEdit("", dialog))
        self.text_length_increment = add_row("Length increment (µm):", pya.QLineEdit("", dialog))
        self.text_element = add_row("Element:", pya.QComboBox(dialog))
        self.text_element.addItems([''] + node_editor_valid_elements)
        self.text_element.editable = True
        self.text_align = add_row("Refpoint names to align to (input,output):", pya.QLineEdit("", dialog))
        self.text_params = add_row("Element parameters:", pya.QPlainTextEdit("", dialog))
        self.button_apply = add_row("", pya.QPushButton("Apply", dialog))
        self.button_apply.clicked(self.update_node_from_form)

    def update_node_from_form(self):
        if self.selection is None:
            return

        try:
            new_node = node_from_text(self.text_x.text, self.text_y.text, self.text_element.currentText,
                                      self.text_inst_name.text, self.text_angle.text,
                                      self.text_length_before.text, self.text_length_increment.text,
                                      self.text_align.text, self.text_params.toPlainText())
        except ValueError as e:
            pya.MessageBox.warning(type(e).__name__, str(e), pya.MessageBox.Ok)
            return

        self.manager.transaction("Edit node")
        replace_node(self.selection['instance'], self.selection['node_index'], new_node)
        self.manager.commit()
        self.selection['node'] = new_node
        self.update()

    def update_form_from_node(self, node):
        (self.text_x.text,
         self.text_y.text,
         self.text_element.currentText,
         self.text_inst_name.text,
         self.text_angle.text,
         self.text_length_before.text,
         self.text_length_increment.text,
         self.text_align.text,
         self.text_params.plainText) = node_to_text(node)

    def deselect(self):
        self.selection = None
        self.last_dialog_position = (self.dialog.x, self.dialog.y)
        self.dialog.hide()

    def select(self, instance, node_index, node):
        if self.selection is not None:
            self.deselect()

        marker = pya.Marker(self.view)
        position = instance.dcplx_trans * node.position
        size = pya.DVector(self.capture_range, self.capture_range) / self.view.viewport_trans().mag
        marker.set(pya.DBox(position - size, position + size))
        marker.vertex_size = 0
        marker.line_style = 2

        self.update_form_from_node(node)
        if self.last_dialog_position is not None:
            self.dialog.move(*self.last_dialog_position)
        self.dialog.show()

        self.selection = {
            'instance': instance,
            'instance_trans': instance.dcplx_trans,
            'node_index': node_index,
            'node': node,
            'marker': marker,
        }

    def activated(self):
        self.is_active = True

    def deactivated(self):
        self.deselect()
        self.is_active = False

    def mouse_click_event(self, p, buttons, prio):
        if prio and buttons == pya.ButtonState.LeftButton and self.is_active:
            cell_view = self.view.active_cellview()
            node_data = get_nodes_near_position(cell_view.cell, p, self.capture_range / self.view.viewport_trans().mag)
            if len(node_data) == 1:
                wg_inst, node, node_index = node_data[0]
                self.select(wg_inst, node_index, node)
            else:
                self.deselect()
            return True
        else:
            return False

    def mouse_moved_event(self, p, _, __):
        # Store current mouse position for use in tracking_position
        self.last_mouse_position = p

    def has_tracking_position(self):  # pylint: disable=no-self-use
        # We provide a custom tracking position only if a node is selected
        return self.selection is not None

    def tracking_position(self):
        if self.selection is not None:
            position = self.selection['instance_trans'].inverted() * self.last_mouse_position
            return position
        else:
            return pya.DPoint(0, 0)  # Only reached if something in the internal state is broken

    def update(self):
        # Update marker size to stay constant in pixels
        if self.selection is not None:
            marker = self.selection['marker']
            position = self.selection['instance_trans'] * self.selection['node'].position
            size = pya.DVector(self.capture_range, self.capture_range) / self.view.viewport_trans().mag
            marker.set(pya.DBox(position - size, position + size))


class EditNodePluginFactory(pya.PluginFactory):
    def __init__(self):
        if pya.Application.instance().is_editable():
            icon_path = os.path.join(os.path.dirname(__file__), 'edit_node_plugin.png')
            self.register(1000, "kqc_edit_node", "Edit Node", icon_path)

            # Set tooltip to something more helpful
            toolbar_action = pya.MainWindow.instance().menu().action('@toolbar.kqc_edit_node')
            if toolbar_action:
                toolbar_action.tool_tip = "Edit individual Node properties of WaveguideComposite elements"

    def create_plugin(self, manager, _, view):  # pylint: disable=no-self-use
        return EditNodePlugin(manager, view)



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import argparse
import logging
import sys
import subprocess
from pathlib import Path
from kqcircuits.simulations.export.export_and_run import export_and_run
from kqcircuits.defaults import TMP_PATH, SCRIPTS_PATH

logging.basicConfig(level=logging.WARN, stream=sys.stdout)

def run():
    parser = argparse.ArgumentParser(description='KQC console scripts')

    subparser = parser.add_subparsers(dest="command")

    simulate_parser = subparser.add_parser('simulate', help='KQC simulation \
            export and run script. Run and export with a single command!')
    mask_parser = subparser.add_parser('mask', help='Build KQC Mask.')

    simulate_parser.add_argument('export_script', type=str, help='Name of the export script')
    simulate_parser.add_argument('--export-path-basename', type=str, default=None,
                                 help='The export folder will be `TMP_PATH / Path(export_path_basename)`, \
                                       if not given, then it will be `TMP_PATH / Path(args.export_script).stem`')
    simulate_parser.add_argument('-q', '--quiet', action="store_true", help='Quiet mode: No GUI')

    mask_parser.add_argument('mask_script', type=str, help='Name of the mask script')
    mask_parser.add_argument('-d', '--debug', action="store_true", help="Debug mode. Use a single process and "
                             "print logs to standard output too.")
    mask_parser.add_argument('-c N', action="store_true", help="Limit the number of used CPUs to 'N'")

    args, args_for_script = parser.parse_known_args()

    if args.command == "simulate":
        script_file = Path(args.export_script)
        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "simulations" / args.export_script)
            if not script_file.is_file():
                logging.error(f"Export script not found at {args.export_script} or {script_file}")
                return
        if args.export_path_basename is not None:
            export_path = TMP_PATH / args.export_path_basename
        else:
            export_path = TMP_PATH / Path(args.export_script).stem
        export_and_run(script_file, export_path, args.quiet, args_for_script)
    elif args.command == "mask":
        script_file = Path(args.mask_script)
        if not script_file.is_file():
            script_file = Path(SCRIPTS_PATH / "masks" / args.mask_script)
            if not script_file.is_file():
                logging.error(f"Mask script not found at {args.mask_script} or {script_file}")
                return
        if args.debug:
            subprocess.call([sys.executable, script_file, '-d'])
        else:  # Windows needs this for multiprocessing
            with open(script_file) as mask_file:
                exec(mask_file.read())  # pylint: disable=exec-used
    else:
        parser.print_usage()



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import pi, tan, degrees, atan2, sqrt

from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridges.airbridge_multi_face import AirbridgeMultiFace
from kqcircuits.elements.element import Element
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_rf import FlipChipConnectorRf
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_curved import WaveguideCoplanarCurved
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import vector_length_and_direction, is_clockwise, get_angle
from kqcircuits.util.parameters import Param, pdt, add_parameters_from


@add_parameters_from(AirbridgeMultiFace)
@add_parameters_from(WaveguideCoplanar, "term1", "term2")
class SpiralResonatorPolygon(Element):
    """The PCell declaration for a polygon shaped spiral resonator.

    The resonator waveguide starts at the first point of `self.input_path` and goes through each each of its points. The
    last point of `self.input_path` will connect to the first point of `self.poly_path`, unless the length of
    `self.input_path` is already longer than `self.length`.

    The polygon shape is defined by `self.poly_path`, where each point is a vertex of the polygon. The points should
    either be in clockwise or counter-clockwise order. The waveguide will first continue along the polygon edges, and
    then spiral inside the polygon such that each segment will be parallel to one of the edges defined by
    `self.polygon_path`.

    Note, that if you want the segment from `self.input_path` to `self.poly_path` to be "a part of the spiral"
    (following the polygon shape), you will need to choose the points such that the edge
    `self.input_path[-1] - self.poly_path[0]` is parallel to the edge `self.poly_path[0] - self.poly_path[-1]`.

    The spacing between waveguides in the spiral can either be chosen manually or automatically. The automatic spacing
    attempts to find the largest possible spacing.

    Airbridges can optionally be placed with a given spacing along the resonator waveguide by setting non-zero value to
    `bridge_spacing`. Alternatively, one can independently set number of airbridges on each straight segment of spiral
    polygon by using parameter `n_bridges_pattern`.

    Non-negative value to `connector_dist` inserts face-to-face connector to spiral resonator so that the beginning and
    the end of resonator will be on different faces.
    """

    length = Param(pdt.TypeDouble, "Resonator length", 5000, unit="μm")
    input_path = Param(pdt.TypeShape, "Input waveguide path", pya.DPath([pya.DPoint(-200, 0), pya.DPoint(0, 0)], 10))
    poly_path = Param(pdt.TypeShape, "Polygon path",
                      pya.DPath([pya.DPoint(0, 800), pya.DPoint(1000, 0), pya.DPoint(0, -800)], 10))
    auto_spacing = Param(pdt.TypeBoolean, "Use automatic spacing", True)
    manual_spacing = Param(pdt.TypeList, "Manual spacing pattern", [300], unit="[μm]")
    bridge_spacing = Param(pdt.TypeDouble, "Airbridge spacing", 0, unit="μm")
    n_bridges_pattern = Param(pdt.TypeList, "Pattern for number of airbridges on edges", [0])
    connector_dist = Param(pdt.TypeDouble, "Face to face connector distance from beginning", -1, unit="µm",
                           docstring="Negative value means single face resonator without connector.")

    def build(self):
        if isinstance(self.input_path, list):
            self.input_path = pya.DPath(self.input_path, 1)
        if isinstance(self.poly_path, list):
            self.poly_path = pya.DPath(self.poly_path, 1)

        if self.auto_spacing:
            self._produce_resonator_automatic_spacing()
        else:
            self._produce_resonator_manual_spacing()

    def _produce_resonator_automatic_spacing(self):
        """Produces polygon spiral resonator with automatically determined waveguide spacing.

        This creates resonators with different spacing, until it finds the largest spacing that can be used to create
        a valid resonator. Only the final resonator with optimal spacing is inserted to `self.cell` in the end.
        """
        def polygon_min_diameter(path):
            p = list(path.each_point())
            n = len(p)
            diams = []
            for i in range(n):
                _, v = vector_length_and_direction(p[(i + 1) % n] - p[i])
                diams.append(max([abs(v.vprod(p[j % n] - p[i])) for j in range(i+2, i+n)]))
            return min(diams)

        # find optimal spacing using bisection method
        min_spacing, max_spacing = 0, polygon_min_diameter(self.poly_path) / 2
        optimal_points = self._produce_path_points([min_spacing])
        if optimal_points is None:
            self.raise_error_on_cell("Cannot create a resonator with the given parameters. Try decreasing the turn "
                                     "radius.", (self.input_path.bbox() + self.poly_path.bbox()).center())

        spacing_tolerance = 0.001
        while max_spacing - min_spacing > spacing_tolerance:
            spacing = (min_spacing + max_spacing) / 2
            points = self._produce_path_points([spacing])
            if points is not None:
                optimal_points = points
                min_spacing = spacing
            else:
                max_spacing = spacing
        self._produce_resonator(optimal_points)
        self.add_port("a", optimal_points[0], optimal_points[0] - optimal_points[1])

    def _produce_resonator_manual_spacing(self):
        """Produces polygon spiral resonator with spacing defined by `self.manual_spacing`.

        If the resonator cannot be created with the chosen spacing, it will instead raise a ValueError.
        """
        sp = [float(s) for s in self.manual_spacing] if isinstance(self.manual_spacing, list) else [self.manual_spacing]
        points = self._produce_path_points(sp)
        if points is None:
            self.raise_error_on_cell("Cannot create a resonator with the given parameters. Try decreasing the spacings "
                                     "or the turn radius.", (self.input_path.bbox() + self.poly_path.bbox()).center())
        self._produce_resonator(points)
        self.add_port("a", points[0], points[0] - points[1])

    def _produce_path_points(self, spacing):
        """Creates resonator path points with the given spacing.
        Function _produce_resonator takes these points as an argument.
        If spacing is unsuitable for creating points, the return value is None.

        Args:
            spacing: spacing between waveguide centers inside the polygon

        Returns:
            List of DPoints or None
        """

        def get_updated_length(pts, prev_length):
            """Updates the resonator length by adding the last point.

            Args:
                pts: list of DPoints
                prev_length: resonator length before adding the last point

            Returns:
                 resonator length including all points (or None if waveguide bends can't fit)
            """
            if len(pts) <= 1:
                return 0.0

            last_segment_len = (pts[-1] - pts[-2]).length()
            if len(pts) == 2:
                return last_segment_len

            # compute new length for the resonator
            _, _, alpha1, alpha2, _ = WaveguideCoplanar.get_corner_data(pts[-3], pts[-2], pts[-1], self.r)
            abs_curve = pi - abs(pi - abs(alpha2 - alpha1))
            corner_cut_dist = self.r * tan(abs_curve / 2)
            updated_length = prev_length - 2 * corner_cut_dist + self.r * abs_curve + last_segment_len

            # if the new segment is not long enough for the curve in the beginning, resonator cannot be created
            if last_segment_len < corner_cut_dist - 1e-5:
                return None

            # if the previous segment is not long enough for the curves at each end, resonator cannot be created
            if len(pts) > 3:
                corner_cut_dist += self._corner_cut_distance(pts[-4], pts[-3], pts[-2])[0]
            if (pts[-2] - pts[-3]).length() < corner_cut_dist - 1e-5:
                return None

            return updated_length

        # segments based on input_path points
        input_points = list(self.input_path.each_point())
        length = 0.0
        points = []
        for ip in input_points:
            # Update length after adding a point
            points.append(ip)
            length = get_updated_length(points, length)
            if length is None:
                return None

            # Test if the resonator is long enough
            if length >= self.length:
                return points

        # segments based on poly_path points
        poly_points = list(self.poly_path.each_point())
        n_poly_points = len(poly_points)
        if n_poly_points > 2:
            poly_edges = [pya.DEdge(poly_points[i], poly_points[(i+1) % n_poly_points]) for i in range(n_poly_points)]
            clockwise = is_clockwise(poly_points)
            # get the normal vectors (toward inside of polygon) of each edge
            normals = []
            for edge in poly_edges:
                _, direction = vector_length_and_direction(edge.p2 - edge.p1)
                normals.append(pya.DVector(direction.y, -direction.x) if clockwise else
                               pya.DVector(-direction.y, direction.x))
            # define amount of spacing for the first round
            shifts = [0.0] * len(poly_edges)
            if len(points) > 0:
                _, input_dir = vector_length_and_direction(poly_points[0] - points[-1])
                _, poly_dir = vector_length_and_direction(poly_points[0] - poly_points[-1])
                shifts[-1] = max(0.0, spacing[-1] * input_dir.sprod(poly_dir))
            i = 0
            current_edge = poly_edges[-1]
            while True:
                # get the edge shifted with spacing from the corresponding poly edge
                next_edge_without_shift = poly_edges[i % n_poly_points]
                shift = shifts[i % n_poly_points] * normals[i % n_poly_points]
                shifts[i % n_poly_points] += spacing[i % len(spacing)]
                next_edge = pya.DEdge(next_edge_without_shift.p1 + shift, next_edge_without_shift.p2 + shift)
                # Find intersection point of the lines defined by current_edge and next_edge.
                # We use `extended` since we want intersections between the "infinite" lines instead of finite edges.
                # Use  "intersection_point = current_edge.cut_point(next_edge)" after Klayout 0.26 support is dropped.
                intersection_point = current_edge.extended(1e7).crossing_point(next_edge.extended(1e7))
                # if the shift was so large that the new segment would end up in the opposite direction,
                # resonator cannot be created
                if i > 0 >= (intersection_point - points[-1]).sprod_sign(current_edge.d()):
                    return None

                # Append point and update length
                points.append(intersection_point)
                length = get_updated_length(points, length)
                if length is None:
                    return None

                # Test if the resonator is long enough
                if length >= self.length:
                    if i < n_poly_points:  # Outest segments don't need overlapping consideration
                        return points

                    # Check outer curve shortcut length to avoid inner segments overlapping with the outer curve.
                    i_out = len(points) - n_poly_points - 1
                    if i_out <= 0:  # Outer curve doesn't exist because input_path is empty
                        return points

                    corner_diff = points[-1] - points[i_out]  # vector from outer corner to inner corner
                    _, inner_dir = vector_length_and_direction(points[-1] - points[-2])
                    s_cut = corner_diff.sprod(inner_dir)
                    if s_cut >= 0:  # For concave corner, segment cannot overlap with outer curve
                        return points

                    # For convex corner, allow straight segment until the outer curve begins.
                    _, outer_dir = vector_length_and_direction(points[i_out] - points[i_out - 1])
                    r_cut, _ = self._corner_cut_distance(points[i_out - 1], points[i_out], points[i_out + 1])
                    if length - max(0.0, s_cut + r_cut * outer_dir.sprod(inner_dir)) >= self.length:
                        return points

                # prepare for the next iteration
                current_edge = next_edge
                i += 1

        return None

    def _produce_resonator(self, points):
        """Produces a polygon spiral resonator with the given path points

        Args:
            points: List of DPoints created by function _produce_path_points
        """
        tmp_cell = self.add_element(WaveguideCoplanar, path=points)
        length = tmp_cell.length()

        # handle correctly the last waveguide segment
        last_segment_curved = self._fix_waveguide_end(points, length)
        term2 = 0 if last_segment_curved else self.term2

        # produce bridges
        self._produce_airbridges(points)

        # insert waveguide with or without connector
        if self.connector_dist >= 0:
            self._produce_wg_with_connector(points, term2)
        else:
            self.insert_cell(WaveguideCoplanar, path=points, term2=term2)

    def _fix_waveguide_end(self, points, current_length):
        """Modifies the last points and places a WaveguideCoplanarCurved element at the end if needed.

        This is required since WaveguideCoplanar cannot end in the middle of a curved segment.

        Args:
            points: list of points used to create the resonator waveguide, may be modified by this method
            current_length: length of the resonator if points are not modified

        Returns:
            True if the last segment is curved and False if it's straight
        """
        extra_len = current_length - self.length
        last_seg_len, last_seg_dir = vector_length_and_direction(points[-1] - points[-2])
        if len(points) > 2:
            v1, v2, alpha1, alpha2, corner_pos = WaveguideCoplanar.get_corner_data(points[-3], points[-2], points[-1],
                                                                                   self.r)
            # distance between points[-2] and start of the curve
            corner_cut_dist = self.r * tan((pi - abs(pi - abs(alpha2 - alpha1))) / 2)
            # check if last waveguide segment is too short to be straight
            if last_seg_len - extra_len < corner_cut_dist:
                # remove last point and move the new last point to the start position of the old curve
                points.pop()
                _, new_last_dir = vector_length_and_direction(points[-1] - points[-2])
                points[-1] -= corner_cut_dist * new_last_dir
                # calculate how long the new curve piece needs to be
                if len(points) > 2:
                    tmp_cell = self.add_element(WaveguideCoplanar, path=points)
                    curve_length = self.length - tmp_cell.length()
                    if curve_length <= 0.0:
                        points[-1] += curve_length * new_last_dir
                        return False
                else:
                    curve_length = self.length - (points[-1] - points[-2]).length()
                curve_alpha = curve_length / self.r
                # add new curve piece at the waveguide end
                fids = [0, 1] if self.connector_dist < 0 else [1, 0]
                curve_cell = self.add_element(WaveguideCoplanarCurved, alpha=curve_alpha,
                                              face_ids=[self.face_ids[f] for f in fids])
                curve_trans = pya.DCplxTrans(1, degrees(alpha1) - v1.vprod_sign(v2)*90, v1.vprod_sign(v2) < 0,
                                             corner_pos)
                self.insert_cell(curve_cell, curve_trans)
                WaveguideCoplanarCurved.produce_curve_termination(self, curve_alpha, self.term2, curve_trans, *fids)
                return True

        # set last point to correct position based on length
        points[-1] = points[-1] - extra_len * last_seg_dir
        return False

    def _produce_airbridges(self, points):
        """Produces airbridges defined either by self.bridge_spacing or self.n_bridges_pattern.

        All airbridges have at least half bridge width distance to end of the straight.

        Args:
            points: list of points used to create the resonator waveguide
        """
        # Create airbridges by self.bridge_spacing
        bridge_width = Airbridge.get_schema()["bridge_width"].default
        if self.bridge_spacing > 0.0:
            dist_to_next_bridge = self.bridge_spacing
            for i in range(0, len(points) - 1):
                segment_len, segment_dir = vector_length_and_direction(points[i + 1] - points[i])
                cut_dist, curve_len = (self._corner_cut_distance(points[i], points[i + 1], points[i + 2])
                                       if i + 2 < len(points) else (0.0, 0.0))
                end_of_straight = segment_len - bridge_width - cut_dist

                angle = degrees(atan2(segment_dir.y, segment_dir.x))
                while dist_to_next_bridge < end_of_straight:
                    pos = points[i] + dist_to_next_bridge * segment_dir
                    self.insert_cell(Airbridge, pya.DCplxTrans(1, angle, False, pos))
                    dist_to_next_bridge += self.bridge_spacing
                dist_to_next_bridge = max(dist_to_next_bridge - segment_len + 2 * cut_dist - curve_len,
                                          cut_dist + bridge_width)

        # Create airbridges by self.n_bridges_pattern
        nb = [int(n) for n in self.n_bridges_pattern] if isinstance(self.n_bridges_pattern, list) else []
        n_beg = self.input_path.num_points()
        if any(nb) and n_beg < len(points) - 2:
            cut_dist0, _ = self._corner_cut_distance(points[n_beg - 1], points[n_beg], points[n_beg + 1])
            for i in range(n_beg, len(points) - 2):
                segment_len, segment_dir = vector_length_and_direction(points[i+1] - points[i])
                cut_dist1, _ = self._corner_cut_distance(points[i], points[i + 1], points[i + 2])

                n_bridges = nb[(i - n_beg) % len(nb)]
                shift = 0.5 - 0.5 * (((i - n_beg) // len(nb)) % 2)
                ab_dist = (segment_len - cut_dist0 - cut_dist1 - 2 * bridge_width) / (n_bridges - 0.5)
                angle = degrees(atan2(segment_dir.y, segment_dir.x))
                for b in range(n_bridges):
                    pos = points[i] + (cut_dist0 + bridge_width + (b + shift) * ab_dist) * segment_dir
                    self.insert_cell(Airbridge, pya.DCplxTrans(1, angle, False, pos))
                cut_dist0 = cut_dist1

    def _produce_wg_with_connector(self, points, term2):
        """Produces waveguide with face-to-face connector.

        Args:
            points: list of points used to create the resonator waveguide
            term2: end termination
        """
        # add connector cell and get connector length
        conn_cell = self.add_element(FlipChipConnectorRf)
        conn_ref = self.get_refpoints(conn_cell)
        port0 = self.face_ids[0] + "_port"
        port1 = self.face_ids[1] + "_port"
        conn_len, conn_dir = vector_length_and_direction(conn_ref[port1] - conn_ref[port0])

        def insert_wg_with_connector(segment, distance):
            s_len, s_dir = vector_length_and_direction(points[segment + 1] - points[segment])
            b_pos = points[segment] + distance * s_dir
            ang = get_angle(s_dir) - get_angle(conn_dir)
            trans = pya.DCplxTrans(1.0, ang, False, b_pos) * pya.DTrans(-conn_ref[port0])
            t_pos = self.insert_cell(conn_cell, trans=trans)[1][port1]
            if segment == 0 and distance < 1e-3:
                WaveguideCoplanar.produce_end_termination(self, t_pos, b_pos, self.term1)
            else:
                self.insert_cell(WaveguideCoplanar, path=points[:segment + 1] + [b_pos], term2=0)
            if segment + 2 == len(points) and s_len - conn_len - distance < 1e-3:
                WaveguideCoplanar.produce_end_termination(self, b_pos, t_pos, term2, face_index=1, opp_face_index=0)
            else:
                self.insert_cell(WaveguideCoplanar, path=[t_pos] + points[segment + 1:],
                                 term1=0, term2=term2, face_ids=self.face_ids[1::-1])

        last = {}  # parameters for last possible connector position
        prev_len = 0.0
        prev_cut_dist = 0.0
        for i, p in enumerate(points):
            if i + 1 >= len(points):
                if last:
                    insert_wg_with_connector(**last)
                    return
                self.raise_error_on_cell("Face-to-face connector cannot fit.",
                                         (self.input_path.bbox() + self.poly_path.bbox()).center())
            corner_cut_dist, corner_length = ((0.0, 0.0) if i + 2 == len(points) else
                                              self._corner_cut_distance(p, points[i + 1], points[i + 2]))
            segment_len, _ = vector_length_and_direction(points[i + 1] - p)
            straight_len = segment_len - prev_cut_dist - corner_cut_dist
            if conn_len <= straight_len:
                last = {'segment': i, 'distance': segment_len - corner_cut_dist - conn_len}
                dist = self.connector_dist - conn_len / 2 - prev_len + prev_cut_dist
                if dist <= last['distance']:
                    insert_wg_with_connector(i, max(prev_cut_dist, dist))
                    return
            prev_len += straight_len + corner_length
            prev_cut_dist = corner_cut_dist

    def _corner_cut_distance(self, point1, point2, point3):
        """Returns the distance from waveguide path corner to the start of the curve and the curve length.

        Args:
            point1: point before corner
            point2: corner point
            point3: point after corner

        Returns:
            corner cut distance, curve length
        """
        _, _, alpha1, alpha2, _ = WaveguideCoplanar.get_corner_data(point1, point2, point3, self.r)
        abs_curve = pi - abs(pi - abs(alpha2 - alpha1))
        return self.r*tan(abs_curve/2), self.r * abs_curve


def rectangular_parameters(above_space=500, below_space=400, right_space=1000, x_spacing=100, y_spacing=100,
                           bridges_left=False, bridges_bottom=False, bridges_right=False, bridges_top=False,
                           r=Element.get_schema()["r"].default, **kwargs):
    """A utility function to easily produce rectangular spiral resonator (old SpiralResonatorRectangle).

    Args:
        above_space: Space above the input (µm)
        below_space: Space below the input (µm)
        right_space: Space right of the input (µm)
        x_spacing: Spacing between vertical segments (µm)
        y_spacing: Spacing between horizontal segments (µm)
        bridges_left: Crossing airbridges left
        bridges_bottom: Crossing airbridges bottom
        bridges_right: Crossing airbridges right
        bridges_top: Crossing airbridges top
        r: Turn radius (µm)

    Returns:
        dictionary of parameters for SpiralResonatorPolygon
    """
    defaults = {"manual_spacing": [y_spacing, x_spacing], "r": r}
    if above_space == 0:
        params = {"input_path": pya.DPath([], 10),
                  "poly_path": pya.DPath([pya.DPoint(0, above_space), pya.DPoint(right_space, above_space),
                                          pya.DPoint(right_space, -below_space), pya.DPoint(0, -below_space)], 10),
                  "n_bridges_pattern": [bridges_top, bridges_right, bridges_bottom, bridges_left]}
    elif below_space == 0:
        params = {"input_path": pya.DPath([], 10),
                  "poly_path": pya.DPath([pya.DPoint(0, -below_space), pya.DPoint(right_space, -below_space),
                                          pya.DPoint(right_space, above_space), pya.DPoint(0, above_space)], 10),
                  "n_bridges_pattern": [bridges_bottom, bridges_right, bridges_top, bridges_left]}
    elif above_space > below_space:
        x1 = sqrt(above_space / (4 * r - above_space)) * r if above_space < 2 * r else r
        x2 = (sqrt((4 * r - above_space) * above_space) if above_space < 2 * r else 2 * r) - x1
        params = {"input_path": pya.DPath([pya.DPoint(0, 0), pya.DPoint(x1, 0)], 10),
                  "poly_path": pya.DPath([pya.DPoint(x2, above_space), pya.DPoint(right_space, above_space),
                                          pya.DPoint(right_space, -below_space), pya.DPoint(x2, -below_space)], 10),
                  "n_bridges_pattern": [bridges_top, bridges_right, bridges_bottom, bridges_left]}
    else:
        x1 = sqrt(below_space / (4 * r - below_space)) * r if below_space < 2 * r else r
        x2 = (sqrt((4 * r - below_space) * below_space) if below_space < 2 * r else 2 * r) - x1
        params = {"input_path": pya.DPath([pya.DPoint(0, 0), pya.DPoint(x1, 0)], 10),
                  "poly_path": pya.DPath([pya.DPoint(x2, -below_space), pya.DPoint(right_space, -below_space),
                                          pya.DPoint(right_space, above_space), pya.DPoint(x2, above_space)], 10),
                  "n_bridges_pattern": [bridges_bottom, bridges_right, bridges_top, bridges_left]}

    return {**defaults, **params, **kwargs}



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from

@add_parameters_from(Chip, frames_enabled=[0, 1])
class DaisyWoven(Chip):
    """Base PCell declaration for a Daisy Woven chip.

    Includes texts in pixel corners, dicing edge, launchers and manually-drawn daisy pattern.
    No input parameters on this class.
    """

    name_chip = Param(pdt.TypeString, "Name of the chip", "DC")

    def build(self):
        self._produce_daisy_face("Daisy_woven")

    def _produce_daisy_face(self, cell_name):
        # first create chip frame to change polarity of manual drawing
        super().produce_structures()

        # import daisy bottom cell
        daisy_cell = Element.create_cell_from_shape(self.layout, cell_name)

        # copy features for both faces
        for face_id in [0, 1]:
            box = pya.DPolygon(self.get_box(face_id))

            # create box
            x_min = min(self.box.p1.x, self.box.p2.x)
            x_max = max(self.box.p1.x, self.box.p2.x)
            y_min = min(self.box.p1.y, self.box.p2.y)
            y_max = max(self.box.p1.y, self.box.p2.y)

            # shorthand notation
            origin_offset_x = 1e3 * (x_max - x_min) / 2.
            origin_offset_y = 1e3 * (y_max - y_min) / 2.

            chip_region = pya.Region([box.to_itype(self.layout.dbu)])  # this is already the shape of the box

            # Using a static file, so use static layer indices
            daisy_shapes_base_metal_gap_wo_grid  = daisy_cell.shapes(self.layout.layer(11 + face_id * 30, 1))
            daisy_shapes_underbump_metallization = daisy_cell.shapes(self.layout.layer(32 + face_id * 30, 4))
            daisy_shapes_indium_bump             = daisy_cell.shapes(self.layout.layer(33 + face_id * 30, 4))

            protection = pya.Region(
                self.cell.begin_shapes_rec(self.get_layer("ground_grid_avoidance", face_id))).merged()
            self.cell.shapes(self.get_layer("ground_grid_avoidance", face_id)).insert(chip_region)

            # extract the bottom Nb layer
            pattern = pya.Region(daisy_shapes_base_metal_gap_wo_grid).moved(
                origin_offset_x, origin_offset_y)
            difference = chip_region - pattern - protection

            # copy design cell layers manually to DaisyWoven cell
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", face_id)).insert(difference)
            self.cell.shapes(self.get_layer("underbump_metallization", face_id)).insert(
                pya.Region(daisy_shapes_underbump_metallization).moved(
                    origin_offset_x, origin_offset_y))
            self.cell.shapes(self.get_layer("indium_bump", face_id)).insert(
                pya.Region(daisy_shapes_indium_bump).moved(
                    origin_offset_x, origin_offset_y))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya


def polygon_with_hsym(points):
    """Polygon with horizontal symmetry.

    Arguments:
        points: List of points to copied to the other side of the symmetry axis. Points at the symmetry axis will
            be doubled.

    Returns:
        DPolygon
    """
    return polygon_with_sym(points, pya.DTrans.M0)


def polygon_with_vsym(points):
    """Polygon with vertical symmetry.

    Arguments:
        points: List of points to copied to the other side of the symmetry axis. Points at the symmetry axis will
            be doubled.

    Returns:
        DPolygon
    """
    return polygon_with_sym(points, pya.DTrans.M90)


def polygon_with_sym(points, mirror_trans):
    """Polygon with symmetry with respect to an axis.

    Arguments:
        points: List of points to copied to the other side of the symmetry axis. Points at the symmetry axis will
            be doubled.
        mirror_trans: pya.DTrans that mirrors the points with respect to an axis

    Returns:
        DPolygon
    """
    # mirror the points and return polygon
    antipoints = [mirror_trans * p for p in reversed(points)]

    return pya.DPolygon([*points, *antipoints])



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from pathlib import Path
import subprocess
import platform
import sys
import logging

logging.basicConfig(level=logging.WARN, stream=sys.stdout)

def export_and_run(export_script: Path, export_path: Path, quiet: bool=False, args=None):
    """
    Exports and runs a KQC simulation.

    Args:
        export_script(Path): path to the simulation export script
        export_path(Path): path where simulation files are exported
        quiet(bool): if True all the GUI dialogs are shown, otherwise not.
        args(list): a list of strings describing arguments to be passed to the simulation script

    Returns:
        a tuple containing

            * export_script(Path): path to the simulation export script
            * export_path(Path): path where simulation files are exported

    """

    if args is None:
        args = []
    else:
        if '--simulation-export-path' in args:
            logging.error("--simulation-export-path is not allowed!")
            sys.exit()

    subprocess.call([sys.executable, export_script,
        '--simulation-export-path', str(export_path)] + args + (['-q'] if quiet else []))


    if (export_path / 'simulation.sh').is_file():
        simulation_shell_script = 'simulation.sh'
    else:
        simulation_shell_script = 'simulation.bat'

    if platform.system() == 'Windows':  # Windows
        subprocess.call(simulation_shell_script, shell=True, cwd=str(export_path))
    elif platform.system() == 'Darwin':  # macOS
        subprocess.call(['bash', simulation_shell_script], cwd=str(export_path))
    else:  # Linux
        subprocess.call(['bash', simulation_shell_script], cwd=str(export_path))

    return export_script, export_path
